#!/usr/bin/env python3
#
# pshub - A Pub/Sub framework implemented in Python
# Copyright (C) 2016  caizixian, lwher
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
from pshub.networking.protocol import parse_stream, prepare_stream, \
    make_message


class PubProtocol(asyncio.Protocol):
    def __init__(self, loop, msg_gen):
        self.loop = loop
        self.msg_gen = msg_gen
        self.rest = bytearray()
        self.count = 0

    def publish_message(self):
        msg = self.msg_gen.next(self.loop)
        logging.debug("Publishing message: {}".format(msg))
        self.transport.write(prepare_stream(make_message('pub', msg)))

    def connection_made(self, transport):
        self.transport = transport
        self.publish_message()

    def data_received(self, data):
        msgs, rest = parse_stream(self.rest, data)
        self.rest = rest
        for ty, body in msgs:
            if ty == 'rep':
                if body["succeeded"]:
                    self.count += 1
                    logging.info(
                        "Successfully published {} messages".format(self.count))
                else:
                    logging.warning("A previous publishing failed.")
            else:
                logging.warning(
                    "Invalid message type: {} from server".format(ty))

        self.publish_message()



#!/usr/bin/env python3
#
# pshub - A Pub/Sub framework implemented in Python
# Copyright (C) 2016  caizixian, lwher
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from operator import lt, le, gt, ge, eq, ne, contains
from functools import wraps
import logging
from collections import defaultdict


def _my_not(f):
    @wraps(f)
    def foo(*args, **kwargs):
        return not f(*args, **kwargs)

    return foo


def _two_partial(func, second_param):
    def fun(first_param):
        return func(first_param, second_param)

    return fun


class RuleFactory:
    """
    RuleFactory provides methods to create Rules
    """
    operators = {
        "contains": contains,
        "<": lt,
        "<=": le,
        ">": gt,
        ">=": ge,
        "=": eq,
        "!=": ne,
        "excludes": _my_not(contains)
    }

    def __init__(self):
        pass

    @staticmethod
    def from_dict(rule):
        """
        Example:
        {"level": {">=": 1, "<": 10}, "message_body": {"contains": "ERROR"}}
        :param rule: rule represented in a Python dictionary
        :return: Rule object
        """
        constraints = defaultdict(list)

        for key, value in rule.items():
            for operator, operand in value.items():
                fun = _two_partial(RuleFactory.operators[operator], operand)
                constraints[key].append(fun)

        return Rule(constraints=constraints)


class Rule:
    def __init__(self, constraints):
        self.constraints = constraints

    def match(self, msg):
        """
        Match message against some constraints
        Example:
        {"level": 2, "message_body": "ERROR occurs"}
        :param msg: message represented in a Python dictionary
        :return: Successfully matched or not
        """
        logging.debug("Matching {}".format(msg))
        for k, v in msg.items():
            for fun in self.constraints[k]:
                if not fun(v):
                    logging.debug("Doesn't match")
                    return False
        logging.debug("Matches")
        return True



#!/usr/bin/env python3
#
# pshub - A Pub/Sub framework implemented in Python
# Copyright (C) 2016  caizixian, lwher
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

delimiter = b"\r\n"


def make_body(obj):
    """

    :param obj: message body, should be JSON-serializable
    :return: encoded message body
    """
    return json.dumps(obj)


def parse_body(string):
    """

    :param string: encoded message body, should be JSON-deserializable
    :return: message body
    """
    return json.loads(string)


def make_message(ty, obj):
    """

    :param ty: type of the message
    :param obj: message body, should be JSON-serializable
    :return: encoded message
    """
    types = {
        'sub': lambda x: 'sub\t' + make_body(x),
        'pub': lambda x: 'pub\t' + make_body(x),
        'rep': lambda x: 'rep\t' + make_body(x)
    }
    return types[ty](obj)


def parse_message(string):
    ty, body = string.split('\t')
    return ty, parse_body(body)


def prepare_stream(msg):
    """

    :param msg: encoded message
    :return:  properly wrapped message for transmission
    """
    return msg.encode('utf-8') + delimiter


def parse_stream(rest, data):
    """

    :param data: currently received data
    :param rest: unprocessed data from last transmission
    :return: message entities extracted from data, unprocessed data
    """
    msgs = []
    tokens = (rest + data).split(delimiter)
    for x in tokens[:-1]:
        msgs.append(parse_message(x.decode('utf-8')))
    return msgs, tokens[-1]



#!/usr/bin/env python3
#
# pshub - A Pub/Sub framework implemented in Python
# Copyright (C) 2016  caizixian, lwher
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
from pshub.networking.protocol import parse_stream, prepare_stream, \
    make_message
from pshub.matchingengine import RuleFactory


class HubClients(object):
    def __init__(self):
        self.subscribers = {}

    def add_subscriber(self, peername, transport, rule):
        self.subscribers[peername] = (transport, rule)

    def publish(self, msg):
        for transport, rule in self.subscribers.values():
            if rule.match(msg):
                transport.write(prepare_stream(make_message("pub", msg)))


class HubProtocol(asyncio.Protocol):
    def __init__(self, loop, clients=None):
        self.loop = loop
        if clients:
            self.clients = clients
        else:
            self.clients = HubClients()
        self.rest = bytearray()

    def connection_made(self, transport):
        self.transport = transport
        self.peername = transport.get_extra_info('peername')
        logging.info("Connection from {}".format(self.peername))

    def data_received(self, data):
        msgs, rest = parse_stream(self.rest, data)
        self.rest = rest
        for ty, body in msgs:
            if ty == "sub":
                self.clients.add_subscriber(self.peername, self.transport,
                                            RuleFactory.from_dict(body))
                self.transport.write(
                    prepare_stream(make_message("rep", {"succeeded": True})))

            elif ty == "pub":
                self.clients.publish(body)
                self.transport.write(
                    prepare_stream(make_message("rep", {"succeeded": True})))
            else:
                logging.warning("Invalid message type: {} from {}".format(ty,
                                                                          self.peername))



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pshub',

    version='0.0.1',

    description='A Pub/Sub framework implemented in Python',
    long_description=long_description,

    url='https://github.com/caizixian/pshub',

    author='Ivan Cai, lwher',
    author_email='caizixian@users.noreply.github.com, 751276936@qq.com',

    license='AGPLv3',

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],

    keywords='asyncio pub sub',

    packages=find_packages(exclude=['examples', 'tests']),

    extras_require={
        'dev': ['pytest'],
        'test': ['pytest'],
    }
)
